/* Copyright (C) 2003 David Fries <david@fries.net>
 * This program is free software; you can redistribute it and/or modify it
 * under the terms of the GNU General Public License as published by the Free
 * Software Foundation
 * $Id: ppmaverage.cc,v 1.12 2006/09/24 22:10:57 david Exp $
 */

#include <sys/types.h>
#include <sys/stat.h>
#include <fcntl.h>
#include <unistd.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <errno.h>
#include <dlfcn.h>
#include <ctype.h>

#include <set>
#include <string>

#include "OpenEXRWriter.h"

#define READERROR(file) \
{\
	fprintf(stderr, "Error reading %s: %s\n", file, strerror(errno));\
	exit(1);\
}

// preserve unique comments from the source images to the target image
typedef std::set<std::string> SetType;

/* Read one integer value from input placing any comment lines in the comment
 * set.  Returns a negative value if there was an error reading the file
 * or there was an error reading the integer.
 */
int get_value(int fd, SetType &comment)
{
	char c;
	std::string s;
	do {
		if(read(fd, &c, 1)!=1)
			return -1;
	} while(isspace(c));

	// process comments
	if(c=='#')
	{
		s=c;
		for(;;)
		{
			if(read(fd, &c, 1)!=1)
				return -1;
			if(c=='\n' || c=='\r')
				break;
			s+=c;
		}
		// If the string is unique this will succeed, otherwise
		// the set already has the comment and we don't need to
		// store another copy.
		comment.insert(s);
		// recursively call this function to return a value or an
		// error.
		return get_value(fd, comment);
	}
	
	s=c;
	do {
		if(read(fd, &c, 1)!=1)
			return -1;
		s+=c;
	} while(!isspace(c));
	// If sscanf fails we still return a negative number
	int value=-2;
	sscanf(s.c_str(), "%d", &value);
	return value;
}

/* Open the file and read the header information, verify that it is a
 * ppm format image, then fill out the width and height as well as place
 * the file pointer at the beginning of the data.
 * http://netpbm.sourceforge.net/doc/ppm.html for a description of the header.
 */
void ReadHeader(const char * filename, int &fd, SetType &comment,
	int &width, int &height)
{
	fd = open(filename, O_RDONLY);
	if(fd == -1)
	{
		fprintf(stderr, "Error opening file %s: %s\n", filename,
			strerror(errno));
		exit(1);
	}
	char magic[2];
	if(read(fd, magic, 2)!=2)
		READERROR(filename);
	if(magic[0]!='P' || magic[1]!='6')
	{
		fprintf(stderr, "%s isn't a ppm file (doesn't have 'P6' magic "
			"value), read %c%c\n", filename, magic[0], magic[1]);
		exit(1);
	}
	width=get_value(fd, comment);
	height=get_value(fd, comment);
	if(width<=0 || height<=0)
	{
		fprintf(stderr, "Bad image size %dx%d for %s\n",
			width, height, filename);
		exit(1);
	}
	int max_value=get_value(fd, comment);
	if(max_value!=255)
	{
		fprintf(stderr, "Requires 255 maximum value per pixel, "
			"found %d\n", max_value);
		exit(1);
	}
}

/* Write a PPM format image from the given data which is a sum of 
 * incount images.  image must be large enough to hold the final image,
 * the memory is presumed to be around and not needed at this point in the
 * program so no need to allocate separate memory.
 */
void WritePPMImage(const char * outfile, const SetType &comment,
	unsigned int *data, unsigned char *image, int incount,
	int width, int height)
{
	// open and then write out the new image
	int result, fd, total=width*height*3;
	if(strcmp(outfile, "-"))
		fd = open(outfile, O_WRONLY|O_CREAT|O_EXCL, 0644);
	else
		fd = 1;
	if(fd == -1)
	{
		fprintf(stderr, "Error opening output image %s: %s\n",
			outfile, strerror(errno));
		exit(1);
	}
	FILE *outstream = fdopen(fd, "w");
	if(!outstream)
	{
		fprintf(stderr, "Error creating file output stream\n");
		exit(1);
	}
	// magic, width, height
	fprintf(outstream, "P6\n%d %d\n", width, height);
	// comments
	SetType::const_iterator iter=comment.begin(), i_end=comment.end();
	for(; iter!=i_end; ++iter)
	{
		fprintf(outstream, "%s\n", iter->c_str());
	}
	// maximum value
	fprintf(outstream, "255\n");
	fflush(outstream);
	result = write(fd, image, total);
	if(result == -1)
	{
		fprintf(stderr, "Error writing file %s: %s\n", outfile,
			strerror(errno));
		exit(1);
	}
	if(result != total)
	{
		fprintf(stderr, "Short write to %s\n", outfile);
		exit(1);
	}
	fclose(outstream);
}

WriteEXRImageFuncType LoadEXRFunction(const char * programname)
{
	#define OPENEXRLIBRARY "libOpenEXRWriter.so"
	const char * library_name = OPENEXRLIBRARY;
	void *library = dlopen(library_name, RTLD_NOW);
	char * errorMessage;
	if(!library)
	{
		errorMessage = strdup(dlerror());
		const char * absName = COMPILE_DIRECTORY "/" OPENEXRLIBRARY;
		library = dlopen(absName, RTLD_NOW);
		if(!library)
		{
			fprintf(stderr, "Error loading required library "
				"%s\n", errorMessage);
			fprintf(stderr, "Also failed when using %s\n",
				dlerror());
			exit(2);
		}
		free(errorMessage);
	}
	const char *symbolName = "WriteEXRImage";
	WriteEXRImageFuncType function = (WriteEXRImageFuncType)
		dlsym(library, symbolName);
	errorMessage = dlerror();
	if(errorMessage)
	{
		fprintf(stderr, "Error reading symbol %s: %s\n", symbolName,
			errorMessage);
		exit(1);
	}
	return function;
	#undef OPENEXRLIBRARY
}

void Usage(const char * program)
{
	fprintf(stderr, "Usage: %s infile1.ppm [ ... infilen.ppm ] "
		"-o [ outfile.ppm | outfile.exr | - ]\n"
		"\t.exr is the OpenEXR format developed by\n"
		"\tIndustrial Light & Magic, supporting 32 bit floating\n"
		"\tpoint values per color\n"
		"\tAll input files must have the same width and height\n"
		"\tThis program averages the color values to correct for\n"
		"\tnoise so any motion is a bad idea.  '-o -' writes a ppm\n"
		"\tfile to stdout.\n", program);
	exit(1);
}

int main(int argc, char ** argv)
{
	if(argc < 4)
		Usage(argv[0]);
	const char *infile[argc], *outfile=NULL;
	enum OutputFormat { IMAGE_PPM, IMAGE_EXR } OutputImage = IMAGE_PPM;
	WriteEXRImageFuncType WriteEXRImageFunction = NULL;
	// read in the file names
	for(int i=1; i<argc; i++)
	{
		if(!strcmp("-o", argv[i]))
		{
			outfile = argv[i+1];
			if(i+1>=argc || i+2<argc)
			{
				fprintf(stderr, "Error, -o can only be listed "
					"once and be the second to last "
					"argument\n");
				Usage(argv[0]);
			}
			// look for the file extension
			const char * ext = outfile;
			while(*ext)
			{
				if(*ext++ == '.')
				{
					if(!strcmp(ext, "ppm"))
						OutputImage = IMAGE_PPM;
					else if(!strcmp(ext, "exr"))
					{
						OutputImage = IMAGE_EXR;
						WriteEXRImageFunction =
							LoadEXRFunction(
								argv[0]);
					}
					else
					{
						fprintf(stderr, "Unknown image "
							"format %s\n", ext);
						exit(1);
					}
				}
			}
			break;
		}
		infile[i-1] = argv[i];
	}
	if(!outfile)
		Usage(argv[0]);
	int width=0, height=0;
	int incount = argc-3;
	unsigned int *data = NULL;
	unsigned char *image = NULL;

	int fd, tmpWidth, tmpHeight, result;
	int total=0;
	SetType comment;
	for( int i=0; i<incount; i++)
	{
		ReadHeader(infile[i], fd, comment, tmpWidth, tmpHeight);
		if(!data)
		{
			// allocate the data after opening the first image
			// since we now know the image size
			width = tmpWidth;
			height = tmpHeight;
			total = width*height*3;
			data = new unsigned int[total];
			// start the average at zero
			memset(data, 0, total*sizeof(unsigned int));
			image = new unsigned char[total];
		}
		if(width!=tmpWidth || height!=tmpHeight)
		{
			fprintf(stderr, "All images must be the same width "
				"and height, %s is %dx%d, expected %dx%d\n",
				infile[i], tmpWidth, tmpHeight, width, height);
			exit(1);
		}
		// read in the image
		result = read(fd, image, total);
		if(result == -1)
			READERROR(infile[i]);
		if(result != total)
		{
			fprintf(stderr, "Short read of %s\n", infile[i]);
			exit(1);
		}

		// add this image to the cululated values
		for(int p=0; p<total; p++)
			data[p] += image[p];

		close(fd);
	}
	// calculate the average image
	for(int p=0; p<total; p++)
		image[p] = (unsigned char)(data[p]/incount);
	
	switch(OutputImage)
	{
	case IMAGE_PPM:
		WritePPMImage(outfile, comment, data, image, incount,
			width, height);
		break;
	case IMAGE_EXR:
		WriteEXRImageFunction(outfile, data, incount, width, height);
		break;
	default:
		fprintf(stderr, "Error unknown image file type\n");
	}

	delete[]image;
	delete[]data;
	
	return 0;
}
